from nmigen import *
from nmigen.cli import main


from functools import reduce


# We know this is not really likely to synthesize well, but we implement here so
# we can get a gold-standard/reference implementation, against which we can test
# our further optimizations, including compression and non-standard representations.

# TODO PLAN:
# 1) get this working in simulation
# 2) make it synthesizable (so it can use FPGA block RAM with reasonable bit-width)
# 3) make it optimized

class LRTable(Elaboratable):
    def __init__(self, number_of_states, number_of_terminals, number_of_reduce_rules, input_array):
        # Parameters

        # So the LR table is a two-dimensional array, where one of the dimensions (rows)
        # is indexed by state number, and the other dimension (columns) is indexed by
        # terminal.

        # Each element of the array can be one of four values:
        # 1) Shift terminal
        # 2) Accept parse
        # 3) Error
        # 4) Reduce with rule number k

        # This means that the length of each element has to be:
        # TIW = (2 bits + number of bits to index the reduce rules)

        # Because this is not initially meant for synthesis efficiency, we
        # initially just choose the width of the memory to be:

        # width = number_of_terminals * TIW

        # and the depth to be:
        # depth = number_of_states

        # We calculate these now:

        reduce_rule_signal = Signal(range(number_of_reduce_rules))
        reduce_rule_bits = len(reduce_rule_signal)
        self.table_entry_width = (reduce_rule_bits + 2)
        self.table_width = number_of_terminals * self.table_entry_width
        self.table_depth = number_of_states


        # Interfaces

        # Input data:
        self.state_in    = Signal(range(number_of_states))
        self.terminal_in = Signal(range(number_of_terminals))

        # Control
        self.in_valid      = Signal(1)

        # Output data:

        self.table_entry_out = Signal(self.table_entry_width)


        # Output control
        self.table_entry_out_valid = Signal(1)


        # Prepare the table for consumption
        assert(len(input_array) == number_of_states)

        rasterized = []
        for row in input_array:
            assert(len(row) == number_of_terminals)
            rasterized.extend(row)

        # Memory
        self.mem = Memory(width=self.table_width, depth=self.table_depth, init=rasterized)


    def elaborate(self, platform):
        m = Module()
        m.submodules.rport = rport = (self.mem).read_port()
        m.submodules.wport = wport = (self.mem).write_port()
        row_address = Signal(self.table_depth)

        m.d.comb += row_address.eq(self.state_in)
        m.d.comb += rport.addr.eq(row_address)


        m.d.sync += self.table_entry_out_valid.eq(self.in_valid)

        # Now we calculate the column address:

        registered_column = Signal(len(self.terminal_in))

        m.d.sync += registered_column.eq(self.terminal_in)


        m.d.comb += self.table_entry_out.eq((rport.data).word_select(registered_column, self.table_entry_width))



        return m










class GOTOtable(Elaboratable):
    def __init__(self, number_of_states, number_of_nonterminals, input_array):
        # Parameters

        # So the GOTO table is a two-dimensional array, where one of the dimensions (rows)
        # is indexed by state number, and the other dimension (columns) is indexed by
        # nonterminal.

        # Each element of the array is of the form:

        # [VALID BIT] [STATE NUMBER]

        # where the valid bit is 1 if there is a valid mapping, and it is zero if there is
        # no valid mapping. Naturally we can optimize this in later implementations, but here
        # we accept an extra bit to keep everything simple

        # This means that the length of each element has to be:
        # TIW = (1 bit + number of bits to index the nonterminals)

        # Because this is not initially meant for synthesis efficiency, we
        # initially just choose the width of the memory to be:

        # width = number_of_terminals * TIW

        # and the depth to be:
        # depth = number_of_states

        # We calculate these now:

        nonterminal_signal = Signal(range(number_of_nonterminals))
        nonterminal_bits = len(nonterminal_signal)
        self.table_entry_width = (nonterminal_bits + 1)
        self.table_width = number_of_nonterminals * self.table_entry_width
        self.table_depth = number_of_states


        # Interfaces

        # Input data:
        self.state_in    = Signal(range(number_of_states))
        self.nonterminal_in = Signal(range(number_of_nonterminals))

        # Control
        self.in_valid      = Signal(1)

        # Output data:

        self.table_entry_out = Signal(self.table_entry_width)


        # Output control
        self.table_entry_out_valid = Signal(1)


        # Prepare the table for consumption
        assert(len(input_array) == number_of_states)

        rasterized = []
        for row in input_array:
            assert(len(row) == number_of_nonterminals)
            rasterized.extend(row)

        # Memory

        self.mem = Memory(width=self.table_width, depth=self.table_depth, init=rasterized)


    def elaborate(self, platform):
        m = Module()
        m.submodules.rport = rport = (self.mem).read_port()
        m.submodules.wport = wport = (self.mem).write_port()
        row_address = Signal(self.table_depth)

        m.d.comb += row_address.eq(self.state_in)
        m.d.comb += rport.addr.eq(row_address)


        m.d.sync += self.table_entry_out_valid.eq(self.in_valid)

        # Now we calculate the column address:

        registered_column = Signal(len(self.nonterminal_in))

        m.d.sync += registered_column.eq(self.nonterminal_in)


        m.d.comb += self.table_entry_out.eq((rport.data).word_select(registered_column, self.table_entry_width))



        return m




class DummyPlug(Elaboratable):

    #def __init__(self):



    def elaborate(self, platform):
        m = Module()

        m.submodules.table = table = GOTOtable(3,2,[[1,2],[3,4],[5,6]])
        counter = Signal(8)
        m.d.sync += counter.eq(counter+1)

        with m.If(counter == 3):
            m.d.comb += table.in_valid.eq(1)
            m.d.comb += table.state_in.eq(1)
            m.d.comb += table.nonterminal_in.eq(1)
        with m.If(counter == 4):
            m.d.comb += table.in_valid.eq(1)
            m.d.comb += table.state_in.eq(2)
        with m.If(counter == 6):
            m.d.comb += table.in_valid.eq(1)
            m.d.comb += table.state_in.eq(0)
            m.d.comb += table.nonterminal_in.eq(1)
        return m



if __name__ == '__main__':
    baka =DummyPlug()
    main(baka)
    #platform.build(DummyPlug())
